@@ -463,3 +463,37 @@ def snap(
         idxs[i] = path[-1]
         dists[i] = d
     return idxs, dists
+
+
+# NOTE: not unit tested
+@njit
+def inflow_idxs(idxs_ds, seq, region):
+    """returns linear indices of most upstream cells within region"""
+    idxs = []
+    mask = np.array([bool(1) for _ in range(idxs_ds.size)])  # all True
+    for idx0 in seq[::-1]:  # up- to downstream
+        idx_ds = idxs_ds[idx0]
+        if idx0 != idx_ds:
+            if mask[idx0] and region[idx_ds] and not region[idx0]:  # in
+                idxs.append(idx0)
+                mask[idx_ds] = False
+            else:
+                mask[idx_ds] = mask[idx0]
+    return np.array(idxs, dtype=idxs_ds.dtype)
+
+
+# NOTE: not unit tested
+@njit
+def outflow_idxs(idxs_ds, seq, region):
+    """returns linear indices of most downstream cells within region"""
+    idxs = []
+    mask = np.array([bool(1) for _ in range(idxs_ds.size)])  # all True
+    for idx0 in seq:  # down- to upstream
+        idx_ds = idxs_ds[idx0]
+        # at mask and region and (pit or out)
+        if mask[idx_ds] and region[idx0] and (idx_ds == idx0 or not region[idx_ds]):
+            idxs.append(idx0)
+            mask[idx0] = False
+        else:
+            mask[idx0] = mask[idx_ds]
+    return np.array(idxs, dtype=idxs_ds.dtype)
